/*****************************************************************************
 ** 
 ** kisdndock.cpp
 ** 
 ** Copyright 2000	by Red Hat GmbH
 **
 ** Author:	 Thomas Woerner (thomas.woerner@redhat.de)
 ** Nearly-rewrite for Red Hat Linux 7.2:
 **          Bernhard "Bero" Rosenkrnzer <bero@redhat.com>
 ** 
 *****************************************************************************/

#include <stdlib.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <qcursor.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kuniqueapp.h>
#include <kpopupmenu.h>
#include <kaction.h>

#include <iostream>
#include "kisdndock.h"
#include "Files.h"

/* XPM */
const char *isdn_xpm[] = {
	"24 24 3 1",
	"		 c None",
	".		c #FFFFFF",
	"+		c #000000",
	"........................",
	".++..++++.+++++..++..++.",
	".++.+++...++.+++.+++.++.",
	".++..+++..++..++.++++++.",
	".++...+++.++.+++.++.+++.",
	".++.++++..+++++..++..++.",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................",
	"........................"
};


/* menu items */
#define MENU_ABOUT		100
#define MENU_STATISTICS 120
#define MENU_CONFIGURE	110
#define MENU_QUIT		 130

QString getValue(QString str)
{
	QString value;
 
	if (str.find('=') < 0)
		return QString("");
 
	// move all before, and '='
	str.remove(0, str.find('=') + 1);
 
	// strip white spaces
	str = str.stripWhiteSpace();
 
	// strip beginning and ending '"'
	if (str.find('"') == 0)
		str.remove(0, 1);
	if (str.findRev('"') == (int) str.length() - 1)
		str.remove(str.length() - 1, 1);
 
	// strip white spaces
	value = str.stripWhiteSpace();
 
	return value;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	KISDNdock
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Type			public constructor
	Options		 (int) argc
				(char **) argv
	Return Value	-none-
	Description	 -none-
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
KISDNdock::KISDNdock() : KMainWindow(0),isdninfo_fd(-1),isdnctrl_fd(-1) {
	dial = hangup = 0;

	// initialize graph
	for (int i=0; i<GRAPH_SIZE; i++)
		graph[i] = 0;

	pixmap = new QPixmap(isdn_xpm);

	resize(pixmap->size());
	painter.begin(pixmap);
	
	// set tooltip
	QToolTip::add(this, i18n("Offline"));

	// create menu and stick it on the button widget
	menu = new KPopupMenu();
	connect(menu, SIGNAL(activated(int)), SLOT(activateMenu(int)));

	// load data and generate menu
	load();

	// create timer for update procedure
	timer = new QTimer(this);
	connect(timer, SIGNAL(timeout()), this, SLOT(update()));

	// call timer every 1000ms
	timer->start(1000, false);
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	~KISDNdock
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Type			public destructor
	Options		 -none-
	Return Value	-none-
	Description	 -none-
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
KISDNdock::~KISDNdock() {
	if (isdninfo_fd >= 0)
		close(isdninfo_fd);
	if (isdnctrl_fd >= 0)
		close(isdnctrl_fd);
	painter.end();
	delete pixmap;
	delete menu;
}

int KISDNdock::info_fd(bool errfatal)
{
	if(isdninfo_fd < 0) {
		// open isdninfo device
		if ((isdninfo_fd = open(DEV_ISDNINFO, O_RDWR)) < 0) {
			if(errfatal) {
				timer->stop();
				KMessageBox::error(0L, i18n("%1: %2").arg(DEV_ISDNINFO).arg(strerror(errno)));
				kapp->quit();
				exit(-1);
			}
		}
	}
	return isdninfo_fd;
}

int KISDNdock::ctrl_fd(bool errfatal)
{
	if (isdnctrl_fd < 0) {
		// open isdnctrl device
		if ((isdnctrl_fd = open(DEV_ISDNCTRL, O_RDWR)) < 0) {
			if(errfatal) {
				timer->stop();
				KMessageBox::error(0L, i18n("%1: %2").arg(DEV_ISDNCTRL).arg(strerror(errno)));
				kapp->quit();
				exit(-1);
			}
		}
	}
	return isdnctrl_fd;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	aboutData
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Type			public
	Options		 -none-
	Return Value	about data
	Description	 -none-
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
const KAboutData *KISDNdock::aboutData() {
	KAboutData *data = new KAboutData("kisdndock", I18N_NOOP("KISDNdock"), "1.0.0",
			 I18N_NOOP("kde isdn dock application"),
			 KAboutData::License_GPL,
			 "(c) 1999-2000 Thomas Wrner, Red Hat GmbH\n(c) 2001 Bernhard \"Bero\" Rosenkrnzer, Red Hat GmbH");
	data->addAuthor("Thomas Wrner", I18N_NOOP("Author"),
			"thomas.woerner@redhat.de");
	data->addAuthor("Bernhard \"Bero\" Rosenkrnzer", I18N_NOOP("Ready for RHL 7.2"), "bero@redhat.com");

	return data;
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	load
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Type			protected
	Options		 -none-
	Return Value	-none-
	Description	 loads provider data from PROC_NET and creates menu
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
void KISDNdock::load() {
	unsigned int i;

	// remove all entries from menu
	menu->clear();
	menu->insertTitle(i18n("KISDNdock"));

	// clear provider data
	provider.clear();

	// make sure IP interfaces are up and running
	QStringList interfaces=Files::glob(IFCFG_PREFIX "ippp*",
			(Files::Types)(Files::File || Files::Link));
	interfaces += Files::glob(IFCFG_PREFIX "isdn*",
			(Files::Types)(Files::File || Files::Link));
	QStringList devList;
	for(QStringList::ConstIterator it=interfaces.begin(); it!=interfaces.end(); it++)
		devList.append((*it).mid(strlen(IFCFG_PREFIX)));
	
	for (unsigned int i=1; i <=devList.count(); i++) {
		QFile file(IFCFG_PREFIX + devList[i-1]);
		bool found = false;
		bool foundmissingname = false;
		QString name;
		QString device;
		QString fallback;
		if (file.open(IO_ReadOnly)) {
			QTextStream ts(&file);
			while (!ts.eof()) {
				QString s = ts.readLine();
				if (s.at(1) == '#')
					continue;
				else if (s.contains("NAME=")) {
					device = devList[i-1];
					name = getValue(s);
					found = true;
					break;
				} else if (s.contains("PROVIDER=")) {
					device = devList[i-1];
					fallback = getValue(s);
					foundmissingname = true;
				} else if (s.contains("DEVICE=") && fallback.isEmpty()) {
					device = devList[i-1];
					fallback = getValue(s);
					foundmissingname = true;
				}
			}
			file.close();
			if (foundmissingname && !found) {
				name = fallback;
				found = true;
			}
			if (found) {
				Provider *pd = new Provider();
				pd->device = device.latin1();
				pd->name = name.latin1();
				provider.append(pd);
			}
		}
	}

	// insert provider in reverse order
	for (i=1; i<=provider.count(); i++) {
		menu->insertItem(i18n("Call %1").arg(provider.at(i-1)->name), i);
	}

	// add remaining items
	menu->insertSeparator();
	KAction *a = actionCollection()->action("help_about_app");
	if (a) menu->insertItem(i18n("&About"), MENU_ABOUT);
	//	menu->insertItem(i18n("&Statistics"), MENU_STATISTICS);
	//	menu->insertItem(i18n("&Configure"), MENU_CONFIGURE);
	//	menu->insertSeparator();
	menu->insertItem(i18n("&Quit"), MENU_QUIT);
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	activateMenu
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Type			private slot - called by menu widget
	Options		 (int) index of menu entry
	Return Value	-none-
	Description	 calls exit or sets dial or hangup variable
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
void KISDNdock::activateMenu(int i) {
	switch (i) {
	case -1:
		break;
	case MENU_ABOUT:
		actionCollection()->action("help_about_app")->activate();
		break;
	case MENU_STATISTICS:
		break;
	case MENU_CONFIGURE:
		break;
	case MENU_QUIT:
		kapp->quit();
		break;
	default:
		if (i >= 0	&&	i <= (int)provider.count()) {
			if (provider.at(i-1)->trying	||	provider.at(i-1)->active)
				hangup = i;
			else
				dial = i;
		}
	}
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	is_trying
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Type			protected
	Options		 (int) index of provider in provider list
	Return Value	(bool) device is connected
	Description	 gets phone number from isdn ioctl if device is connected 
				and sets phone number in provider entry
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
bool KISDNdock::is_trying(int i) {
	isdn_net_ioctl_phone phone;

	// initialize ioctl phone structure
	strcpy(phone.name, provider.at(i-1)->device.latin1());

	// get phonenummer of device if connected
	int fd=info_fd(false);
	if(fd<0)
		return false;
	if (ioctl(fd, IIOCNETGPN, &phone) == 0) {
		if (strlen(phone.phone) > 0) {
			provider.at(i-1)->phone = phone.phone;
			return true;
		}
	}
	return false;
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	is_active
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Type			protected
	Options		 (int) index of provider in provider list
	Return Value	-none-
	Description	 gets phone number from isdn ioctl if device is connected 
				and sets phone number in provider entry
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
bool KISDNdock::is_active(int i) {
	char data[1024];

	QString s;
	QFile file(PROC_ROUTE);
	if (file.open(IO_ReadOnly)) {
		while (file.readLine(data, 1024) >= 0) {
			s = data;
			if (s.contains(provider.at(i-1)->device)) {
				file.close();
				return true;
			}
		}
		file.close();
	}

	return false;
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	is_trying
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Type			protected
	Options		 (int) index of provider in provider list
	Return Value	-none-
	Description	 gets phone number from isdn ioctl if device is connected 
				and sets phone number in provider entry
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
void KISDNdock::get_index(int i) {
	char phone[ISDN_MAX_CHANNELS][30];

	// get all phone numbers (evil)
	char buf[8192];
	if (read(info_fd(), buf, sizeof(buf)) > 0) {
		sscanf(strstr(buf, "phone:"),
			 "phone: %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s",
			 phone[0], phone[1], phone[2], phone[3], phone[4], phone[5],
			 phone[6], phone[7], phone[8], phone[8], phone[10], phone[11],
			 phone[12], phone[13], phone[14], phone[15]);
	}

	// get index from phone array
	for (int j=0; j<ISDN_MAX_CHANNELS; j++) {
		if (strncmp(phone[j], provider.at(i-1)->phone.latin1(),
				strlen(provider.at(i-1)->phone.latin1())) == 0)
		{
			provider.at(i-1)->idx = j;
			return;
		}
	}
	
	provider.at(i-1)->idx = -1;
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	update
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Type			protected slot
	Options			-none-
	Return Value	-none-
	Description	 	updates provider data, dials and hangsup device, gets bytes,
					updates tooltip and draws graph
					function is called every 1000ms
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
void KISDNdock::update() {
	int result;
	Provider *pd;
	unsigned int i;
	bool trying, active;
	QFileInfo fi(PROC_NET);
	QString str, tip;
	static unsigned int last_bytes=0;
	unsigned int bt=0;
	time_t time_now;

	time_now = time(NULL);

	// dial (user request)
	if (dial > 0) {
		// Stop the timer - ifup can take considerably more than a second,
		// causing this to be run twice if the timer isn't stopped.
		timer->stop();
		pd = provider.at(dial-1);
		// Get the interface up
		int ifup=system(QString("/sbin/ifup %1").arg(pd->device.latin1()));
		if(ifup != 0) {
			KMessageBox::sorry(this, i18n("ifup %1 failed. This might indicate a bad ISDN configuration.").arg(pd->device.latin1()));
		}
		// give ifup some time to finish...
		sleep(1);
		// dial
		if ((result = ioctl(ctrl_fd(), IIOCNETDIL, pd->device.latin1())) < 0)
			KMessageBox::sorry(this, i18n("%1: %2").arg(pd->device.latin1()).arg(strerror(errno)));
		/* this will come back
		else
			cerr << pd->device.latin1() << " triggered\n";
		*/

		// Restart the timer
		timer->start(1000, false);

		dial = 0;
	}
	// hangup (user request)
	if (hangup > 0) {
		// Stop the timer - ifdown can take considerably more than a second,
		// causing this to be run twice if the timer isn't stopped.
		timer->stop();

		pd = provider.at(hangup-1);
		// hangup
		if ((result = ioctl(ctrl_fd(), IIOCNETHUP, pd->device.latin1())) < 0)
			KMessageBox::sorry(this, i18n("%1: %2").arg(pd->device.latin1()).arg(strerror(errno)));
		/* this will com back
		else {
			if (result)
				cerr << pd->device.latin1() << " not connected\n";
			else
				cerr << pd->device.latin1() << " hung up\n";
		}
		*/
		// down the interface
		system(QString("/sbin/ifdown %1").arg(pd->device.latin1()));

		// Restart the timer
		timer->start(1000, false);

		hangup = 0;
	}

	// find activated and inactivated providers
	active = trying = false;
	for (i=1; i<=provider.count(); i++) {
		pd = provider.at(i-1);
		if (is_trying(i)) {
			trying = true;
			// connected device
			if (!pd->trying	&&	!pd->active) {
				// update menu entry
				menu->changeItem(i18n("Hangup %1").arg(pd->name), i);
				menu->setItemEnabled(i, true);

				// update status
				pd->trying = true;

				// set time
				pd->time = time_now;

				get_index(i);
			}
			// activated device
			if (!pd->active	&&	is_active(i)) {
				pd->trying = false;
				pd->active = true;
				active = true;
			}
			if (pd->active)
				active = true;
		} else {
			// inactivated or disconnected device
			if (pd->trying	||	pd->active) {
				pd->clearFields();
				menu->changeItem(i18n("Call %1").arg(pd->name), i);
			}
		}
	}

	// get i/o bytes, generate tooltip string, enable/disable menu entries
	if (trying || active) {
		for (i=1; i<=provider.count(); i++) {
			pd = provider.at(i-1);
			if (pd->trying || pd->active) {
				int fd=info_fd(false);
				if(fd>=0) {
					if (ioctl(fd, IIOCGETCPS, &iobytes) == 0) {
						if (pd->idx >= 0 && pd->idx < ISDN_MAX_CHANNELS) {
							pd->ibytes = iobytes[pd->idx].ibytes;
							pd->obytes = iobytes[pd->idx].obytes;

							bt += pd->ibytes;
							bt += pd->obytes;

							// ignore bytes of devices connected at program start
							// and aready transfered at program start
							if (time_now == pd->time && last_bytes == 0 &&
								(pd->ibytes > 8192 || pd->obytes > 8192))
							{
								last_bytes += pd->ibytes;
								last_bytes += pd->obytes;
							}
						}
					}
				}

				// generate tooltip string
				if (tip.length() > 0) // append newline if strlen > 0
					tip.append("\n");
				int t = time_now - pd->time;
				str.sprintf(" (h:m:s) %02d:%02d:%02d, (i/o): %dk/%dk", 
						t/60/60, (t/60)%60, t%60,
						pd->ibytes/1024, pd->obytes/1024);
				tip += pd->name+" "+i18n("online")+" "+str;

				// enable menu entry
				menu->setItemEnabled(i, true);
			} else
				menu->setItemEnabled(i, false); // disable menu entry
		}
	} else {
		for (i=1; i<=provider.count(); i++)
			if (!menu->isItemEnabled(i)) {
				menu->setItemEnabled(i, true); // enable menu entry
				if (tip.length() == 0)
					tip = i18n("Offline");
			}
	}

	// set tooltip
	if (tip.length() > 0)
		QToolTip::add(this, tip);

	// draw icon
	if (active)
		painter.fillRect(0, 7, 24, 24, red);
	else if (trying)
		painter.fillRect(0, 7, 24, 24, yellow);
	else if (provider.count() > 0)
		painter.fillRect(0, 7, 24, 24, green);
	else
		painter.fillRect(0, 7, 24, 24, white);

	// set new graph value
	if (bt < last_bytes)	last_bytes = bt = 0;

	/* draw */

	// shift left i/o byte array
	for (i=0; i<GRAPH_SIZE-1; i++)
		graph[i] = graph[i+1];
	graph[GRAPH_SIZE-1] = bt - last_bytes;

	// get maximum value
	unsigned int max=1024;
	for (i=0; i<GRAPH_SIZE; i++)
		if (max < graph[i])
			max = 1024*((graph[i]+512)/1024);
	float fl_max = 16.0 / max;

	// draw 
	if (max > 1024) {
		painter.setPen(black);
		for (i=1024; i<=max; i+=1024)
			painter.drawLine(0, 23-(int)(fl_max * i),
							 GRAPH_SIZE-1, 23-(int)(fl_max * i));
	}
	// draw graph
	int j;
	painter.setPen(blue);
	for (i=0; i<GRAPH_SIZE; i++) {
		j = (int)(fl_max * graph[i]);
		if (j > 16)
			j = 16;
		if (j > 0)
			painter.drawLine(i, 23, i, 23-j);
	}

	last_bytes = bt;

	paintEvent(NULL);
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	paintEvent
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Type			protected
	Options		 (QPaintEvent *) paint event
	Return Value	-none-
	Description	 copies pixmap to button area
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
void KISDNdock::paintEvent(QPaintEvent *) {
	QPainter p(this);
	int x = (width() - pixmap->width()) / 2;
	int y = (height() - pixmap->height()) / 2;
	if ( x < 0 ) x = 0;
	if ( y < 0 ) y = 0;
	p.drawPixmap(x , y, *pixmap);
	p.end();
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	mousePressEvent
	- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	Type			protected
	Options		 (QMouseEvent *) mouse event
	Return Value	-none-
	Description	 popups menu if mouse button is pressed
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
void KISDNdock::mousePressEvent(QMouseEvent* event) {
	if (event->button() == RightButton || event->button() == LeftButton)
		menu->popup(QCursor::pos());
}
// vim:ts=4:sw=4
